(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Copyright (C) 2003-2007 Milenko Mitrovic                                 *
 *  Mail: dcoder@dsp-worx.de                                                 *
 *  Web:  http://www.dsp-worx.de                                             *
 *                                                                           *
 *  This Program is free software; you can redistribute it and/or modify     *
 *  it under the terms of the GNU General Public License as published by     *
 *  the Free Software Foundation; either version 2, or (at your option)      *
 *  any later version.                                                       *
 *                                                                           *
 *  This Program is distributed in the hope that it will be useful,          *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of           *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             *
 *  GNU General Public License for more details.                             *
 *                                                                           *
 *  You should have received a copy of the GNU General Public License        *
 *  along with GNU Make; see the file COPYING.  If not, write to             *
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.    *
 *  http://www.gnu.org/copyleft/gpl.html                                     *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

unit BassDecoder;

{$I Compiler.inc}

interface

uses
  Windows, SysUtils, SyncObjs, WideStrUtils, Math, DirectSound;

type
  TBassExtension = record
    Extension: WideString;
    IsMOD: Boolean;
    DLL: WideString;
  end;

  HFX = DWORD;          // DX8 effect handle

const
  BASS_DLL: WideString = 'bass.dll';

  BASS_EXTENSIONS_COUNT = 20;
  BASS_EXTENSIONS: array[0..BASS_EXTENSIONS_COUNT -1] of TBassExtension = (
    (Extension: '.aac';  IsMOD: False; DLL: 'bass_aac.dll'),
    (Extension: '.alac'; IsMOD: False; DLL: 'bass_alac.dll'),
    (Extension: '.als';  IsMOD: False; DLL: 'bass_alac.dll'),
    (Extension: '.ape';  IsMOD: False; DLL: 'bass_ape.dll'),
    (Extension: '.flac'; IsMOD: False; DLL: 'bass_flac.dll'),
    (Extension: '.m4a';  IsMOD: False; DLL: 'bass_aac.dll'),
    (Extension: '.mac';  IsMOD: False; DLL: 'bass_ape.dll'),
    (Extension: '.mp3';  IsMOD: False; DLL: 'bass.dll'),
    (Extension: '.ogg';  IsMOD: False; DLL: 'bass.dll'),
    (Extension: '.mpc';  IsMOD: False; DLL: 'bass_mpc.dll'),
    (Extension: '.tta';  IsMOD: False; DLL: 'bass_tta.dll'),
    (Extension: '.wv';   IsMOD: False; DLL: 'bass_wv.dll'),
    (Extension: '.ofr';  IsMOD: False; DLL: 'bass_ofr.dll'),

    (Extension: '.it';   IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.mo3';  IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.mod';  IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.mtm';  IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.s3m';  IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.umx';  IsMOD: True; DLL: 'bass.dll'),
    (Extension: '.xm';   IsMOD: True; DLL: 'bass.dll')
  );

  BASS_PLUGINS_COUNT = 8;
  BASS_PLUGINS: array[0..BASS_PLUGINS_COUNT -1] of WideString = (
    'bass_aac.dll',
    'bass_alac.dll',
    'bass_ape.dll',
    'bass_flac.dll',
    'bass_mpc.dll',
    'bass_tta.dll',
    'bass_ofr.dll',
    'bass_wv.dll'
  );

  BASS_STREAM_DECODE      = $200000;
  BASS_MUSIC_DECODE       = BASS_STREAM_DECODE;
  BASS_MUSIC_RAMP         = $200;
  BASS_MUSIC_POSRESET     = $8000;
  BASS_STREAM_PRESCAN     = $20000;
  BASS_MUSIC_PRESCAN      = BASS_STREAM_PRESCAN;
  BASS_UNICODE            = $80000000;
  BASS_SAMPLE_8BITS       = 1;   // 8 bit
  BASS_SAMPLE_FLOAT       = 256; // 32-bit floating-point
  BASS_SYNC_MIXTIME       = $40000000;
  BASS_SYNC_META          = 4;
  BASS_POS_BYTE           = 0; // byte position

  BASS_CTYPE_STREAM_AAC   = $10B00;    // AAC
  BASS_CTYPE_STREAM_MP4   = $10B01;    // MP4
  BASS_CTYPE_STREAM_MP3   = $10005;
  BASS_CTYPE_STREAM_OGG   = $10002;

  BASS_TAG_HTTP           = 3; // HTTP headers : array of null-terminated strings
  BASS_TAG_ICY            = 4; // ICY headers : array of null-terminated strings

  BASS_CONFIG_NET_BUFFER  = 12;
  BASS_CONFIG_NET_PREBUF  = 15;
  BASS_CONFIG_NET_AGENT   = 16;

  {!!! Fix: New code !!!}
  BASS_FX_DX8_PARAMEQ     = 7;

  BASS_SYNC_OGG_CHANGE    = 12;

  BASS_TAG_META           = 5; // ICY metadata : ANSI string

  BASS_POS_MUSIC_ORDER    = 1; // order.row position, MAKELONG(order,row)
  BASS_POS_DECODE         = $10000000; // flag: get the decoding (not playing) position
  BASS_POS_DECODETO       = $20000000; // flag: decode to the position instead of seeking

  // BASS_StreamGetFilePosition modes
  BASS_FILEPOS_CURRENT    = 0;
  BASS_FILEPOS_DECODE     = BASS_FILEPOS_CURRENT;
  BASS_FILEPOS_DOWNLOAD   = 1;
  BASS_FILEPOS_END        = 2;
  BASS_FILEPOS_START      = 3;
  BASS_FILEPOS_CONNECTED  = 4;
  BASS_FILEPOS_BUFFER     = 5;

  BASS_EQ_BAND_NUM        = 10;
  {!!! Fix: New code !!!}

type
  BASS_CHANNELINFO = record
    freq: DWORD;        // default playback rate
    chans: DWORD;       // channels
    flags: DWORD;       // BASS_SAMPLE/STREAM/MUSIC/SPEAKER flags
    ctype: DWORD;       // type of channel
    origres: DWORD;     // original resolution
    plugin: DWORD;      // plugin
    sample: DWORD;      // sample
    filename: PAnsiChar; // filename
  end;

  {!!! Fix: Old code !!!
  TShoutcastMetaDataCallback = procedure(AText: String) of Object;
  !!! Fix: Old code !!!}
  {!!! Fix: New code !!!}
  TShoutcastMetaDataCallback = procedure(const ATitle,AStream:WideString) of object;

  BASS_DX8_PARAMEQ = record
    fCenter: Single;
    fBandwidth: Single;
    fGain: Single;
  end;

  TBassEqSetings = packed record
    Enabled : LongBool;
    Gains   : array[0..BASS_EQ_BAND_NUM-1] of Single;
  end;
  {!!! Fix: New code !!!}

  TShoutcastBufferCallback   = procedure(ABuffer: PByte; ASize: Integer) of Object;

  TBassDecoder = class
  protected
    BASS_Init:                  function(device: Integer; freq, flags: DWORD; win: HWND; clsid: PGUID): BOOL; stdcall;
    BASS_Free:                  function: BOOL; stdcall;
    BASS_PluginLoad:            function(filename: Pointer; flags: DWORD): Cardinal; stdcall;
    BASS_MusicLoad:             function(mem: BOOL; f: Pointer; offset: Int64; length, flags, freq: DWORD): Cardinal; stdcall;
    BASS_StreamCreateURL:       function(url: Pointer; offset: DWORD; flags: DWORD; proc: Pointer; user: Pointer): Cardinal; stdcall;
    BASS_StreamCreateFile:      function(mem: BOOL; f: Pointer; offset, length: Int64; flags: DWORD): Cardinal; stdcall;
    BASS_MusicFree:             function(handle: Cardinal): BOOL; stdcall;
    BASS_StreamFree:            function(handle: Cardinal): BOOL; stdcall;
    BASS_ChannelGetData:        function(handle: DWORD; buffer: Pointer; length: DWORD): DWORD; stdcall;
    BASS_ChannelGetInfo:        function(handle: DWORD; var info: BASS_CHANNELINFO): BOOL;stdcall;
    BASS_ChannelGetLength:      function(handle, mode: DWORD): Int64; stdcall;
    BASS_ChannelSetPosition:    function(handle: DWORD; pos: Int64; mode: DWORD): BOOL; stdcall;
    BASS_ChannelGetPosition:    function(handle, mode: DWORD): Int64; stdcall;
    BASS_ChannelSetSync:        function(handle: DWORD; type_: DWORD; param: int64; proc: Pointer; user: Pointer): DWORD; stdcall;
    BASS_ChannelRemoveSync:     function(handle: DWORD; sync: DWORD): BOOL; stdcall;
    BASS_ChannelGetTags:        function(handle: DWORD; tags: DWORD): PChar; stdcall;
    BASS_SetConfig:             function(option, value: DWORD): DWORD; stdcall;
    {!!! Fix: New code !!!}
    BASS_GetConfig:             function(option: DWORD): DWORD; stdcall;
    BASS_StreamGetFilePosition: function (handle: DWORD; mode: DWORD): Int64; stdcall;

    BASS_ChannelSetFX: function (handle, type_: DWORD; priority: LongInt): HFX; stdcall;
    BASS_ChannelRemoveFX: function (handle: DWORD; fx: HFX): BOOL; stdcall;
    BASS_FXSetParameters: function (handle: HFX; par: Pointer): BOOL; stdcall;
    BASS_FXGetParameters: function (handle: HFX; par: Pointer): BOOL; stdcall;
    BASS_FXReset: function (handle: HFX): BOOL; stdcall;
    {!!! Fix: New code !!!}

    FMetaDataCallback: TShoutcastMetaDataCallback;
    FBufferCallback: TShoutcastBufferCallback;
    FBuffersizeMS: Integer;
    FPrebufferMS: Integer;

    FLibrary: THandle;
    FOptimFROGDLL: THandle;
    FStream: Cardinal;
    FSync: Cardinal;
    {!!! Fix: New code !!!}
    {
    FSync2: Cardinal;
    }
    FBitrate  : Integer;
    FEqEnabled: Boolean;
    FEqHandles: array[0..10-1] of HFX;
    FEqParams : array[0..10-1] of Single;
    {!!! Fix: New code !!!}

    FIsMOD: Boolean;
    FIsURL: Boolean;

    FChannels: Integer;
    FSampleRate: Integer;
    FBytesPerSample: Integer;
    FFloat: Boolean;
    FMSecConv: Int64;
    FIsShoutcast: Boolean;
    FType: Cardinal;

    procedure LoadBASS;
    procedure UnloadBASS;
    procedure LoadPlugins;

    function GetStreamInfos: Boolean;
    procedure GetHTTPInfos;

    function GetDuration: Int64;
    function GetPosition: Int64;
    {!!! Fix: Old code !!!
    procedure SetPosition(APositionMS: Int64);
     !!! Fix: Old code !!!}
    {!!! Fix: New code !!!}
    function GetIsConnected: Boolean;
    procedure SetEq;
    procedure UpdateEq;
    procedure RemoveEq;
    {!!! Fix: New code !!!}
    function GetExtension: WideString;
  public
    constructor Create(AMetaDataCallback: TShoutcastMetaDataCallback; ABufferCallback: TShoutcastBufferCallback; ABuffersizeMS: Integer; APrebufferMS: Integer);
    destructor Destroy; override;

    {!!! Fix: Old code !!!
    function Load(AFileName: WideString): Boolean;
     !!! Fix: Old code !!!}
    {!!! Fix: New code !!!}
    function Load(AFileName: WideString; AFloatSamples:Boolean): Boolean;
    {!!! Fix: New code !!!}
    procedure Close;
    {!!! Fix: New code !!!}
    function SetPosition(APositionMS: Int64):Boolean;
    {!!! Fix: New code !!!}

    function GetData(ABuffer: Pointer; ASize: Integer): integer;
    {!!! Fix: New code !!!}
    function CheckBufferFilled:Boolean;
    function GetBitrate:Integer;

    function GetEqEnabled(out AEqEnabled:LongBool): HRESULT; stdcall;
    function SetEqEnabled(AEqEnabled:LongBool): HRESULT; stdcall;
    function GetEqBandGain(ABandIx:Integer;out AGain:Single):HRESULT; stdcall;
    function SetEqBandGain(ABandIx:Integer;AGain:Single):HRESULT; stdcall;
    {!!! Fix: New code !!!}
  public
    property DurationMS: Int64 read GetDuration;
    property PositionMS: Int64 read GetPosition {!!! Fix: Old code !!! write SetPosition !!! Fix: Old code !!!};

    property Channels: Integer read FChannels;
    property SampleRate: Integer read FSampleRate;
    property BytesPerSample: Integer read FBytesPerSample;
    property Float: Boolean read FFloat;
    property MSecConv: Int64 read FMSecConv;
    property IsShoutcast: Boolean read FIsShoutcast;
    {!!! Fix: New code !!!}
    property IsURL: Boolean read FIsURL;
    property PrebufferMS:Integer read FPrebufferMS;
    property IsConnected:Boolean read GetIsConnected;
    {!!! Fix: New code !!!}
    property Extension: WideString read GetExtension;
  end;

  function GetFilterDirectory: WideString;

implementation

uses
  BassFilter;

(*** Utilities ****************************************************************)

function WideLastDelimiter(const Delimiters, S: WideString): Integer;
var
  P: PWideChar;
begin
  Result := Length(S);
  P := PWideChar(Delimiters);
  while Result > 0 do
  begin
    if (S[Result] <> #0) and (WStrScan(P, S[Result]) <> nil) then
      Exit;
    Dec(Result);
  end;
end;

function WideExtractFilePath(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := WideLastDelimiter('\:', FileName);
  Result := Copy(FileName, 1, I);
end;

function WideExtractFileExt(const FileName: WideString): WideString;
var
  I: Integer;
begin
  I := WideLastDelimiter('.\:', FileName);
  if (I > 0) and (FileName[I] = '.') then
    Result := Copy(FileName, I, MaxInt) else
    Result := '';
end;

function GetFilterDirectory: WideString;
var
  pathDll: array[0..MAX_PATH -1] of WideChar;
  res: Cardinal;
begin
  res := GetModuleFileNameW(HInstance, pathDll, MAX_PATH);
  if res = 0 then
  begin
    Result := '';
    Exit;
  end;

  Result := WideExtractFilePath(pathDll);
end;

function IsMODFile(AFileName: WideString): Boolean;
var
  ext: WideString;
  i: Integer;
begin
  ext := WideLowerCase(WideExtractFileExt(AFileName));

  for i := 0 to BASS_EXTENSIONS_COUNT -1 do
  begin
    if (BASS_EXTENSIONS[i].Extension = ext) then
    begin
      Result := BASS_EXTENSIONS[i].IsMOD;
      Exit;
    end;
  end;

  Result := False;
end;

{!!! Fix: Old code !!!
function ReplaceICYX(AString: WideString): WideString;
var
  strPos: Integer;
begin
  strPos := Pos(WideString('icyx'), AString);

  if (strPos = 1) and (WStrLen(PWideChar(AString)) > 3)  then
  begin
    AString[1] := 'h';
    AString[2] := 't';
    AString[3] := 't';
    AString[4] := 'p';
  end;

  Result := AString;
end;
!!! Fix: Old code !!!}

{!!! Fix: New code !!!}
function ReplaceICYX(S:WideString):WideString;
begin
  if (Length(S) >= 4) and
     ((S[1] = 'i') or (S[1] = 'I')) and
     ((S[2] = 'c') or (S[2] = 'C')) and
     ((S[3] = 'y') or (S[3] = 'Y')) and
     ((S[4] = 'x') or (S[4] = 'X')) then
  begin
    S[1] := 'h'; S[2] := 't';
    S[3] := 't'; S[4] := 'p';
  end;
  Result := S;
end;
{!!! Fix: New code !!!}

function IsURLPath(AFileName: WideString): Boolean;
begin
  Result := (Pos(WideString('http://'), WideLowerCase(AFileName)) > 0) or (Pos(WideString('ftp://'), WideLowerCase(AFileName)) > 0)
end;

(*** Callbacks ****************************************************************)
{!!! Fix: New code !!!}
function HTMLDecode(const AStr: String): String;
var
  Sp, Rp, Cp, Tp: PChar;
  S: String;
  I, Code: Integer;
begin
  SetLength(Result, Length(AStr));
  Sp := PChar(AStr);
  Rp := PChar(Result);
  while Sp^ <> #0 do
  begin
    case Sp^ of
      '&': begin
             Cp := Sp;
             Inc(Sp);
             case Sp^ of
               'a': if AnsiStrPos(Sp, 'amp;') = Sp then  { do not localize }
                    begin
                      Inc(Sp, 3);
                      Rp^ := '&';
                    end;
               'l',
               'g': if (AnsiStrPos(Sp, 'lt;') = Sp) or (AnsiStrPos(Sp, 'gt;') = Sp) then { do not localize }
                    begin
                      Cp := Sp;
                      Inc(Sp, 2);
                      while (Sp^ <> ';') and (Sp^ <> #0) do
                        Inc(Sp);
                      if Cp^ = 'l' then
                        Rp^ := '<'
                      else
                        Rp^ := '>';
                    end;
               'q': if AnsiStrPos(Sp, 'quot;') = Sp then  { do not localize }
                    begin
                      Inc(Sp,4);
                      Rp^ := '"';
                    end;
               '#': begin
                      Tp := Sp;
                      Inc(Tp);
                      while (Sp^ <> ';') and (Sp^ <> #0) do
                        Inc(Sp);
                      SetString(S, Tp, Sp - Tp);
                      Val(S, I, Code);
                      Rp^ := Chr((I));
                    end;
               else begin
                 Sp := Cp;
                 Rp^ := Sp^;
               end;
             end;
         end
    else
      Rp^ := Sp^;
    end;
    Inc(Rp);
    Inc(Sp);
  end;
  SetLength(Result, Rp - PChar(Result));
end;

procedure OnMetaData(handle: DWORD; channel, data: DWORD; user: DWORD); stdcall;

  function GetStreamAttr(AString:PChar;const AName:string):string;
  var
    tag,ltag: string;
  begin
    Result:='';
    while (AString <> nil) and (AString^ <> #0) do begin
      tag:=Trim(AString); ltag:=AnsiLowerCase(tag);
      if Pos(AName,ltag) = 1 then begin
        Delete(tag, 1, Length(AName));
        Result:=Trim(tag);
        if Result <> '' then Break
      end;
      inc(AString, Length(AString) + 1);
    end;
  end;

var
  pcTags: PChar;
  sTags,lsTags: string;
  Title,Stream: string;
  wTitle,wStream: WideString;
  sBitrate: string;
  idx,BR,C: Integer;
begin
  if Assigned(TBassDecoder(user).FMetaDataCallback) then begin
    Title:=''; Stream:='';
    pcTags:=TBassDecoder(user).BASS_ChannelGetTags(channel,BASS_TAG_ICY);
    if Assigned(pcTags) then
      Stream:=GetStreamAttr(pcTags,'icy-name:');
    if Stream = '' then begin
      pcTags:=TBassDecoder(user).BASS_ChannelGetTags(channel,BASS_TAG_HTTP);
      if Assigned(pcTags) then
        Stream:=GetStreamAttr(pcTags,'icy-name:');
    end;
    TBassDecoder(user).FBitrate:=0;
    if Assigned(pcTags) then begin
      sBitrate:=GetStreamAttr(pcTags,'icy-br:');
      if sBitrate <> '' then begin
        val(sBitrate,BR,C);
        if C > 0 then TBassDecoder(user).FBitrate:=0
        else TBassDecoder(user).FBitrate:=BR*1000;
      end;
    end;

    sTags:=TBassDecoder(user).BASS_ChannelGetTags(channel,BASS_TAG_META);
    if sTags <> '' then begin
      lsTags:=AnsiLowerCase(sTags);
      idx:=Pos('streamtitle=',lsTags);
      if idx > 0 then begin
        // Shoutcast Metadata
        sTags:=Copy(sTags,idx+12,Length(sTags)-idx-12);
        idx:=Pos(';',sTags);
        if idx > 0 then
          Delete(sTags,idx,Length(sTags)-idx+1);
        Title:=AnsiDequotedStr(sTags,'''');
      end else begin
        idx:=Pos('title=',lsTags);
        if idx > 0 then
          Title:=Copy(sTags,idx+6,Length(sTags)-idx-5);
      end;
    end;
    Stream:=HTMLDecode(Stream);
    wStream:=UTF8Decode(Stream);
    if wStream = '' then wStream:=Stream;
    Title:=HTMLDecode(Title);
    wTitle:=UTF8Decode(Title);
    if wTitle = '' then wTitle:=Title;
    TBassDecoder(user).FMetaDataCallback(wTitle,wStream);
  end;
end;
{!!! Fix: New code !!!}

{!!! Fix: New code !!!}
{!!!
procedure OnOggMetaData(handle: DWORD; channel, data: DWORD; user: DWORD); stdcall;
var
  metaStr: String;
  resStr: String;
  idx: Integer;
begin
  if Assigned(TBassDecoder(user).FMetaDataCallback) then
  begin
  end;
end;
!!!}
{!!! Fix: New code !!!}

procedure OnShoutcastData (buffer: PByte; length: DWORD; user: DWORD); stdcall;
begin
  if (buffer <> nil) and Assigned(TBassDecoder(user).FBufferCallback)
    then TBassDecoder(user).FBufferCallback(buffer, length);
end;

(*** TBassDecoder *************************************************************)

constructor TBassDecoder.Create(AMetaDataCallback: TShoutcastMetaDataCallback; ABufferCallback: TShoutcastBufferCallback; ABuffersizeMS: Integer; APrebufferMS: Integer);
var
  path: WideString;
begin
  inherited Create;

  path := path;
  path := GetFilterDirectory + WideString('OptimFROG.dll');
  FOptimFROGDLL := LoadLibraryW(PWideChar(path));

  FMetaDataCallback := AMetaDataCallback;
  FBufferCallback := ABufferCallback;
  FBuffersizeMS := ABuffersizeMS;
  FPrebufferMS := APrebufferMS;

  LoadBASS;
  LoadPlugins;
end;

destructor TBassDecoder.Destroy;
begin
  Close;

  if (FOptimFROGDLL <> 0)
    then FreeLibrary(FOptimFROGDLL);


  inherited Destroy;
end;

procedure TBassDecoder.LoadBASS;
var
  path: WideString;
begin
  path := GetFilterDirectory + BASS_DLL;

  if (GetFileAttributesW(PWideChar(path)) <> Cardinal(-1)) then
  begin
    FLibrary := LoadLibraryW(PWideChar(path));
  end;

  if FLibrary = 0 then
  begin
    Exit;
  end;

  @BASS_Init                  := GetProcAddress(FLibrary, 'BASS_Init');
  @BASS_Free                  := GetProcAddress(FLibrary, 'BASS_Free');
  @BASS_PluginLoad            := GetProcAddress(FLibrary, 'BASS_PluginLoad');
  @BASS_MusicLoad             := GetProcAddress(FLibrary, 'BASS_MusicLoad');
  @BASS_StreamCreateURL       := GetProcAddress(FLibrary, 'BASS_StreamCreateURL');
  @BASS_StreamCreateFile      := GetProcAddress(FLibrary, 'BASS_StreamCreateFile');
  @BASS_MusicFree             := GetProcAddress(FLibrary, 'BASS_MusicFree');
  @BASS_StreamFree            := GetProcAddress(FLibrary, 'BASS_StreamFree');
  @BASS_ChannelGetData        := GetProcAddress(FLibrary, 'BASS_ChannelGetData');
  @BASS_ChannelGetInfo        := GetProcAddress(FLibrary, 'BASS_ChannelGetInfo');
  @BASS_ChannelGetLength      := GetProcAddress(FLibrary, 'BASS_ChannelGetLength');
  @BASS_ChannelSetPosition    := GetProcAddress(FLibrary, 'BASS_ChannelSetPosition');
  @BASS_ChannelGetPosition    := GetProcAddress(FLibrary, 'BASS_ChannelGetPosition');
  @BASS_ChannelSetSync        := GetProcAddress(FLibrary, 'BASS_ChannelSetSync');
  @BASS_ChannelRemoveSync     := GetProcAddress(FLibrary, 'BASS_ChannelRemoveSync');
  @BASS_ChannelGetTags        := GetProcAddress(FLibrary, 'BASS_ChannelGetTags');
  @BASS_SetConfig             := GetProcAddress(FLibrary, 'BASS_SetConfig');
  {!!! Fix: New code !!!}
  @BASS_GetConfig             := GetProcAddress(FLibrary, 'BASS_GetConfig');
  @BASS_StreamGetFilePosition := GetProcAddress(FLibrary, 'BASS_StreamGetFilePosition');
  @BASS_ChannelSetFX          := GetProcAddress(FLibrary, 'BASS_ChannelSetFX');
  @BASS_ChannelRemoveFX       := GetProcAddress(FLibrary, 'BASS_ChannelRemoveFX');
  @BASS_FXSetParameters       := GetProcAddress(FLibrary, 'BASS_FXSetParameters');
  @BASS_FXGetParameters       := GetProcAddress(FLibrary, 'BASS_FXGetParameters');
  @BASS_FXReset               := GetProcAddress(FLibrary, 'BASS_FXReset');
  {!!! Fix: New code !!!}

  BASS_Init(0, 44100, 0, GetDesktopWindow, nil);

  if (FPrebufferMS = 0)
    then FPrebufferMS := 1;

  BASS_SetConfig(BASS_CONFIG_NET_AGENT, Integer(PChar('DC-Bass Source')));
  BASS_SetConfig(BASS_CONFIG_NET_BUFFER, FBuffersizeMS);
  BASS_SetConfig(BASS_CONFIG_NET_PREBUF, FBuffersizeMS * 100 div FPrebufferMS);
end;

procedure TBassDecoder.UnloadBASS;
begin
  if (FLibrary > 0) and Assigned(Bass_Free) then
  begin
    try
      if (InstanceCount = 0) then
      begin
        BASS_Free;
      end;
    except
      // crashes mplayer2.exe ???
    end;

    FreeLibrary(FLibrary);
    FLibrary := 0;

    @BASS_Init                  := nil;
    @BASS_Free                  := nil;
    @BASS_PluginLoad            := nil;
    @BASS_MusicLoad             := nil;
    @BASS_StreamCreateURL       := nil;
    @BASS_StreamCreateFile      := nil;
    @BASS_MusicFree             := nil;
    @BASS_StreamFree            := nil;
    @BASS_ChannelGetData        := nil;
    @BASS_ChannelGetInfo        := nil;
    @BASS_ChannelGetLength      := nil;
    @BASS_ChannelSetPosition    := nil;
    @BASS_ChannelGetPosition    := nil;
    @BASS_ChannelSetSync        := nil;
    @BASS_ChannelRemoveSync     := nil;
    @BASS_ChannelGetTags        := nil;
    @BASS_SetConfig             := nil;
    {!!! Fix: New code !!!}
    @BASS_GetConfig             := nil;
    @BASS_StreamGetFilePosition := nil;
    @BASS_ChannelSetFX          := nil;
    @BASS_ChannelRemoveFX       := nil;
    @BASS_FXSetParameters       := nil;
    @BASS_FXGetParameters       := nil;
    @BASS_FXReset               := nil;
    {!!! Fix: New code !!!}
  end;
end;

procedure TBassDecoder.LoadPlugins;
var
  path: WideString;
  i: Integer;
begin
  if (FLibrary = 0)
    then Exit;

  path := GetFilterDirectory;

  for i := 0 to BASS_PLUGINS_COUNT -1 do
  begin
    BASS_PluginLoad(PWideChar(path + BASS_PLUGINS[i]), BASS_UNICODE);
  end;
end;

{!!! Fix: Old code !!!
function TBassDecoder.Load(AFileName: WideString): Boolean;
!!! Fix: Old code !!!}
{!!! Fix: New code !!!}
function TBassDecoder.Load(AFileName: WideString; AFloatSamples:Boolean): Boolean;
begin
  Close;

  if (FLibrary = 0) then
  begin
    Result := False;
    Exit;
  end;

  {!!! Fix: New code !!!}
  if Assigned(FMetaDataCallback) then
    FMetaDataCallback('','');
  {!!! Fix: New code !!!}

  FIsShoutcast := False;
  AFileName := ReplaceICYX(AFileName);

  FIsMOD := IsMODFile(AFileName);
  FIsURL := IsURLPath(AFileName);

  if FIsMOD then
  begin
    if not FIsURL then
    begin
      FStream := BASS_MusicLoad(False, PWideChar(AFileName), 0, 0, BASS_MUSIC_DECODE or BASS_MUSIC_RAMP or BASS_MUSIC_POSRESET or BASS_MUSIC_PRESCAN or BASS_UNICODE, 0);
    end;
  end else
  begin
    if FIsURL then
    begin
      FStream := BASS_StreamCreateURL(PWideChar(AFileName), 0,
        {!!! Fix: New code !!!}
        Cardinal(Ord(AFloatSamples))*BASS_SAMPLE_FLOAT or
        {!!! Fix: New code !!!}
        BASS_STREAM_DECODE or BASS_UNICODE, @OnShoutcastData, Self);
      {!!! Fix: New code !!!}
      if FStream <> 0 then
      {!!! Fix: New code !!!}
        FSync := BASS_ChannelSetSync(FStream, BASS_SYNC_META, 0, @OnMetaData, Self);
      {!!! Fix: Old code !!!
      FIsShoutcast := GetDuration = 0;
      !!! Fix: Old code !!!}
    end else
    begin
      FStream := BASS_StreamCreateFile(False, PWideChar(AFileName), 0, 0,
        {!!! Fix: New code !!!}
        Cardinal(Ord(AFloatSamples))*BASS_SAMPLE_FLOAT or
        {!!! Fix: New code !!!}
        BASS_STREAM_DECODE or BASS_UNICODE);
    end;
    {!!! Fix: New code !!!}
    {
    if FStream <> 0 then
      FSync2 := BASS_ChannelSetSync(FStream, BASS_SYNC_OGG_CHANGE, 0, @OnOggMetaData, Self);
    }
    {!!! Fix: New code !!!}
  end;

  if (FStream = 0) then
  begin
    Result := False;
    Exit;
  end;

  if not GetStreamInfos then
  begin
    Close;
    Result := False;
    Exit;              
  end;

  Result := True;
  GetBitrate;
end;

procedure TBassDecoder.Close;
begin
  if FStream = 0
    then Exit;

  {!!! Fix: New code !!!}
  RemoveEq;
  {
  if FSync2 <> 0 then
  begin
    BASS_ChannelRemoveSync(FStream, FSync2);
  end;
  }
  FBitrate:=0;
  {!!! Fix: New code !!!}

  if FSync <> 0 then
  begin
    BASS_ChannelRemoveSync(FStream, FSync);
  end;

  if FIsMOD
    then BASS_MusicFree(FStream)
    else BASS_StreamFree(FStream);

  FChannels := 0;
  FSampleRate := 0;
  FBytesPerSample := 0;
  FFloat := False;
  FMSecConv := 0;

  FStream := 0;
end;

function TBassDecoder.GetData(ABuffer: Pointer; ASize: Integer): integer;
begin
  Result := BASS_ChannelGetData(FStream, ABuffer, ASize);
end;

{!!! Fix: New code !!!}
function TBassDecoder.CheckBufferFilled:Boolean;
var
  BP,CP,EP,CD: Int64;
begin
  Result := True;
  if (FStream <> 0) and FIsURL and (FMSecConv > 0) and
     Assigned(@BASS_StreamGetFilePosition) then begin
     BP:=BASS_StreamGetFilePosition(FStream,BASS_FILEPOS_BUFFER);
     EP:=BASS_StreamGetFilePosition(FStream,BASS_FILEPOS_END);
     CD:=DurationMS;
     if (BP < 0) or (EP <= 0) or (CD <= 0) or (BP >= EP) then Exit;
     CP:=BASS_StreamGetFilePosition(FStream,BASS_FILEPOS_CURRENT);
     if CP < 0 then Exit;
     Result:=(BP-CP)*CD div EP > FPrebufferMS;
  end;
end;

function TBassDecoder.GetBitrate:Integer;
var
  BS,BD: Int64;
begin
  Result := 0;
  if FStream <> 0 then begin
     if FBitrate <> 0 then begin
        Result:=FBitrate; Exit;
     end;
     BS:=BASS_StreamGetFilePosition(FStream,BASS_FILEPOS_END);
     if BS = 0 then Exit;
     if FIsShoutcast then
       BD:=BASS_GetConfig(BASS_CONFIG_NET_BUFFER)
     else BD:=DurationMS;
     if BD = 0 then Exit;
     Result:=Round(BS*8000/BD);
     FBitrate:=Result;
  end;
end;

function TBassDecoder.GetEqEnabled(out AEqEnabled:LongBool): HRESULT;
var
  I: Integer;
begin
  AEqEnabled:=FEqEnabled; Result:=S_OK;
  if AEqEnabled and (FStream <> 0) then
    for I:=0 to High(FEqHandles) do
      if FEqHandles[I] = 0 then begin
        Result:=S_FALSE; Exit;
      end;
end;

function TBassDecoder.SetEqEnabled(AEqEnabled:LongBool): HRESULT;
begin
  Result:=S_OK;
  if AEqEnabled = FEqEnabled then Exit;
  FEqEnabled:=AEqEnabled;
  if FEqEnabled then SetEq
  else RemoveEq;
end;

function TBassDecoder.GetEqBandGain(ABandIx:Integer;out AGain:Single):HRESULT;
begin
  if (ABandIx < 0) or (ABandIx > High(FEqParams)) then Result:=E_FAIL
  else begin
    AGain:=FEqParams[ABandIx]; Result:=S_OK;
  end;
end;

function TBassDecoder.SetEqBandGain(ABandIx:Integer;AGain:Single):HRESULT;
begin
  if (ABandIx < 0) or (ABandIx > High(FEqParams)) then Result:=E_FAIL
  else begin
    FEqParams[ABandIx]:=AGain; Result:=S_OK;
    UpdateEq;
  end;
end;

{!!! Fix: New code !!!}

function TBassDecoder.GetStreamInfos: Boolean;
var
  info: BASS_CHANNELINFO;
begin
  if not BASS_ChannelGetInfo(FStream, info) then
  begin
    Result := False;
    Exit;
  end;

  if (info.chans = 0) or (info.freq = 0) then
  begin
    Result := False;
    Exit;
  end;

  FFloat := (info.flags and BASS_SAMPLE_FLOAT) > 0;
  FSampleRate := info.freq;
  FChannels := info.chans;
  FType := info.ctype;

  if FFloat then
  begin
    FBytesPerSample := 4;
  end else
  begin
    if ((info.flags and BASS_SAMPLE_8BITS) > 0) then
    begin
      FBytesPerSample := 1;
    end else
    begin
      FBytesPerSample := 2;
    end;
  end;

  FMSecConv := FSampleRate * FChannels * FBytesPerSample;

  if (FMSecConv = 0) then
  begin
    Result := False;
    Exit;
  end;

  {!!! Fix: New code !!!}
  FIsShoutcast := FIsURL and (GetDuration = 0);
  {!!! Fix: New code !!!}

  GetHTTPInfos;

  Result := True;
end;

{!!! Fix: Old code !!!
procedure TBassDecoder.GetHTTPInfos;

  procedure GetNameTag(AString: PChar);
  var
    tag: String;
  begin
    while (AString <> nil) and (AString^ <> #0) do
    begin
      tag := AString;
      if (Pos('icy-name:', tag) > 0) then
      begin
        Delete(tag, 1, 9);
        tag := Trim(tag);
        if (tag <> '') and Assigned(FMetaDataCallback)
          then FMetaDataCallback(tag);
      end;

      inc(AString, Length(AString) + 1);
    end;
  end;

var
  icyTags: PChar;
  httpHeaders: PChar;
begin
  if not FIsShoutcast
    then Exit;

  httpHeaders := BASS_ChannelGetTags(FStream, BASS_TAG_HTTP);
  if (Assigned(httpHeaders))
    then GetNameTag(httpHeaders);

  icyTags := BASS_ChannelGetTags(FStream, BASS_TAG_ICY);
  if (Assigned(icyTags))
    then GetNameTag(icyTags);
end;
!!! Fix: Old code !!!}

{!!! Fix: New code !!!}
procedure TBassDecoder.GetHTTPInfos;
begin
  if not FIsShoutcast then Exit;
  OnMetaData(0,FStream,0,Cardinal(Self));
end;
{!!! Fix: New code !!!}

function TBassDecoder.GetDuration: Int64;
begin
  if (FMSecConv = 0) then
  begin
    Result := 0;
    Exit;
  end;

  if (FStream = 0) then
  begin
    Result := 0;
    Exit;
  end;

  // bytes = samplerate * channel * bytes_per_second
  // msecs = (bytes * 1000) / (samplerate * channels * bytes_per_second)

  Result := (BASS_ChannelGetLength(FStream, BASS_POS_BYTE) * 1000) div FMSecConv;
end;

function TBassDecoder.GetPosition: Int64;
begin
  if (FMSecConv = 0) then
  begin
    Result := 0;
    Exit;
  end;

  if (FStream = 0) then
  begin
    Result := 0;
    Exit;
  end;

  Result := (BASS_ChannelGetPosition(FStream, BASS_POS_BYTE) * 1000) div FMSecConv;
end;

{!!! Fix: Old code !!!
procedure TBassDecoder.SetPosition(APositionMS: Int64);
var
  pos: Int64;
begin
  if (FMSecConv = 0) then
  begin
    Exit;
  end;

  if (FStream = 0) then
  begin
    Exit;
  end;

  pos := (APositionMS * FMSecConv) div 1000;

  BASS_ChannelSetPosition(FStream, pos, BASS_POS_BYTE);
end;
!!! Fix: Old code !!!}

{!!! Fix: New code !!!}
function TBassDecoder.SetPosition(APositionMS: Int64): Boolean;
var
  pos,cur: Int64;
begin
  Result:=False;

  if (FMSecConv = 0) then
  begin
    Exit;
  end;

  if (FStream = 0) then
  begin
    Exit;
  end;

  pos := (APositionMS * FMSecConv) div 1000;

  Result:=BASS_ChannelSetPosition(FStream, pos, BASS_POS_BYTE);

  if not Result and FIsURL and not IsShoutcast and
     (APositionMS > PositionMS) then begin
    cur:=BASS_ChannelGetPosition(FStream, BASS_POS_BYTE);
    repeat
      inc(cur,(100* FMSecConv) div 1000);
      Result:=BASS_ChannelSetPosition(FStream, cur, BASS_POS_BYTE);
    until (cur >= pos) or not Result;
  end;
end;

function TBassDecoder.GetIsConnected: Boolean;
begin
  Result:=(FStream <> 0) and FIsURL and Assigned(@BASS_StreamGetFilePosition) and
    (BASS_StreamGetFilePosition(FStream,BASS_FILEPOS_CONNECTED) = 1);
end;

procedure TBassDecoder.SetEq;
var
  I  : Integer;
begin
  if FStream = 0 then Exit;
  RemoveEq;
  for I := 0 to High(FEqHandles) do
    FEqHandles[I]:=BASS_ChannelSetFX(FStream,BASS_FX_DX8_PARAMEQ,1);
  UpdateEq;
end;

procedure TBassDecoder.UpdateEq;
var
  I  : Integer;
  BW : Double;
  EP : BASS_DX8_PARAMEQ;
begin
  if (FStream = 0) or not FEqEnabled then Exit;
  BW:=Power(DSFXPARAMEQ_CENTER_MAX/DSFXPARAMEQ_CENTER_MIN,1/High(FEqHandles));
  EP.fCenter:=DSFXPARAMEQ_CENTER_MIN; EP.fBandwidth:=BW*6;
  for I := 0 to High(FEqHandles) do begin
    EP.fGain:=FEqParams[I];
    if FEqHandles[I] <> 0 then
      BASS_FXSetParameters(FEqHandles[I],@EP);
    EP.fCenter:=EP.fCenter*BW;
  end;
end;

procedure TBassDecoder.RemoveEq;
var
  I: Integer;
begin
  if FStream <> 0 then
    for I := 0 to High(FEqHandles) do
      if FEqHandles[I] <> 0 then begin
        BASS_ChannelRemoveFX(FStream,FEqHandles[I]);
        FEqHandles[I]:=0;
      end;
end;

{!!! Fix: New code !!!}

function TBassDecoder.GetExtension: WideString;
begin
  case FType of
    BASS_CTYPE_STREAM_AAC:  Result := 'aac';
    BASS_CTYPE_STREAM_MP4:  Result := 'mp4';
    BASS_CTYPE_STREAM_MP3:  Result := 'mp3';
    BASS_CTYPE_STREAM_OGG:  Result := 'ogg';
    else                    Result := 'mp3';
  end;
end;

end.
